@@ -23,13 +23,14 @@ enum intel_options {
 	OPTION_TRIPLE_BUFFER,
 	OPTION_PREFER_OVERLAY,
 	OPTION_HOTPLUG,
-	OPTION_RELAXED_FENCING,
+	OPTION_REPROBE,
 #if defined(XvMCExtension) && defined(ENABLE_XVMC)
 	OPTION_XVMC,
 #define INTEL_XVMC 1
 #endif
 #ifdef USE_SNA
 	OPTION_ZAPHOD,
+	OPTION_VIRTUAL,
 	OPTION_TEAR_FREE,
 	OPTION_CRTC_PIXMAPS,
 #endif
