@@ -192,15 +192,12 @@ sna_set_fallback_mode(ScrnInfoPtr scrn)
 #endif
 }
 
-static Bool sna_become_master(struct sna *sna)
+static Bool sna_set_desired_mode(struct sna *sna)
 {
 	ScrnInfoPtr scrn = sna->scrn;
 
 	DBG(("%s\n", __FUNCTION__));
 
-	if (intel_get_master(scrn))
-		return FALSE;
-
 	if (!xf86SetDesiredModes(scrn)) {
 		xf86DrvMsg(scrn->scrnIndex, X_WARNING,
 			   "failed to restore desired modes on VT switch\n");
@@ -220,6 +217,8 @@ static Bool sna_become_master(struct sna *sna)
 static Bool sna_create_screen_resources(ScreenPtr screen)
 {
 	struct sna *sna = to_sna_from_screen(screen);
+	PixmapPtr new_front;
+	unsigned hint;
 
 	DBG(("%s(%dx%d@%d)\n", __FUNCTION__,
 	     screen->width, screen->height, screen->rootDepth));
@@ -227,17 +226,22 @@ static Bool sna_create_screen_resources(ScreenPtr screen)
 	assert(sna->scrn == xf86ScreenToScrn(screen));
 	assert(sna->scrn->pScreen == screen);
 
+	/* free the data used during miInitScreen */
 	free(screen->devPrivate);
 	screen->devPrivate = NULL;
 
 	sna_accel_create(sna);
 
-	sna->front = screen->CreatePixmap(screen,
-					  screen->width,
-					  screen->height,
-					  screen->rootDepth,
-					  SNA_CREATE_FB);
-	if (!sna->front) {
+	hint = SNA_CREATE_FB;
+	if (sna->flags & SNA_IS_HOSTED)
+		hint = 0;
+
+	new_front = screen->CreatePixmap(screen,
+					 screen->width,
+					 screen->height,
+					 screen->rootDepth,
+					 hint);
+	if (!new_front) {
 		xf86DrvMsg(screen->myNum, X_ERROR,
 			   "[intel] Unable to create front buffer %dx%d at depth %d\n",
 			   screen->width,
@@ -247,34 +251,42 @@ static Bool sna_create_screen_resources(ScreenPtr screen)
 		return FALSE;
 	}
 
-	if (!sna_pixmap_force_to_gpu(sna->front, MOVE_WRITE)) {
+	if (!sna_pixmap_force_to_gpu(new_front, MOVE_WRITE)) {
 		xf86DrvMsg(screen->myNum, X_ERROR,
 			   "[intel] Failed to allocate video resources for front buffer %dx%d at depth %d\n",
 			   screen->width,
 			   screen->height,
 			   screen->rootDepth);
-		goto cleanup_front;
+		screen->DestroyPixmap(new_front);
+		return FALSE;
 	}
 
-	screen->SetScreenPixmap(sna->front);
+	screen->SetScreenPixmap(new_front);
+	assert(screen->GetScreenPixmap(screen) == new_front);
+	assert(sna->front == new_front);
+	screen->DestroyPixmap(new_front); /* transfer ownership to screen */
 
-	/* Only preserve the fbcon, not any subsequent server regens */
-	if (serverGeneration == 1)
-		sna_copy_fbcon(sna);
-
-	if (!sna_become_master(sna)) {
+	if (intel_get_master(sna->scrn)) {
 		xf86DrvMsg(screen->myNum, X_ERROR,
 			   "[intel] Failed to become DRM master\n");
-		goto cleanup_front;
+		screen->DestroyPixmap(sna->front);
+		sna->front = NULL;
+		return FALSE;
+	}
+
+	/* Only preserve the fbcon, not any subsequent server regens */
+	if (serverGeneration == 1 && (sna->flags & SNA_IS_HOSTED) == 0)
+		sna_copy_fbcon(sna);
+
+	if (!sna_set_desired_mode(sna)) {
+		xf86DrvMsg(screen->myNum, X_ERROR,
+			   "[intel] Failed to set initial mode\n");
+		screen->DestroyPixmap(sna->front);
+		sna->front = NULL;
+		return FALSE;
 	}
 
 	return TRUE;
-
-cleanup_front:
-	screen->SetScreenPixmap(NULL);
-	screen->DestroyPixmap(sna->front);
-	sna->front = NULL;
-	return FALSE;
 }
 
 static void sna_selftest(void)
@@ -287,7 +299,7 @@ static bool has_pageflipping(struct sna *sna)
 	drm_i915_getparam_t gp;
 	int v;
 
-	if (sna->flags & SNA_NO_WAIT)
+	if (sna->flags & (SNA_IS_HOSTED | SNA_NO_WAIT))
 		return false;
 
 	v = 0;
@@ -413,6 +425,7 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 	sna = to_sna(scrn);
 	sna->scrn = scrn;
 	sna->pEnt = pEnt;
+	sna->flags = 0;
 
 	scrn->displayWidth = 640;	/* default it */
 
@@ -425,13 +438,28 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 	fd = intel_get_device(scrn);
 	if (fd == -1) {
 		xf86DrvMsg(scrn->scrnIndex, X_ERROR,
-			   "Failed to become DRM master.\n");
+			   "Failed to claim DRM device.\n");
 		goto cleanup;
 	}
 
-	preferred_depth = sna->info->gen < 030 ? 15 : 24;
-	if (!fb_supports_depth(fd, preferred_depth))
+	if (!sna_xmir_create(sna))
+		goto cleanup;
+
+	/* Sanity check */
+	if (hosted() && (sna->flags & SNA_IS_HOSTED) == 0) {
+		xf86DrvMsg(scrn->scrnIndex, X_ERROR,
+			   "Failed to setup hosted device.\n");
+		goto cleanup;
+	}
+
+	if (sna->xmir) {
+		/* XXX query depth from xmir */
 		preferred_depth = 24;
+	} else {
+		preferred_depth = sna->info->gen < 030 ? 15 : 24;
+		if (!fb_supports_depth(fd, preferred_depth))
+			preferred_depth = 24;
+	}
 
 	if (!xf86SetDepthBpp(scrn, preferred_depth, 0, 0,
 			     Support32bppFb |
@@ -444,7 +472,8 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 	case 16:
 	case 24:
 	case 30:
-		if (fb_supports_depth(fd, scrn->depth))
+		if ((sna->flags & SNA_IS_HOSTED) ||
+		    fb_supports_depth(fd, scrn->depth))
 			break;
 	default:
 		xf86DrvMsg(scrn->scrnIndex, X_ERROR,
@@ -467,6 +496,9 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 
 	intel_detect_chipset(scrn, sna->pEnt, sna->PciInfo);
 
+	if (!sna_xmir_pre_init(sna))
+		goto cleanup;
+
 	kgem_init(&sna->kgem, fd, sna->PciInfo, sna->info->gen);
 	if (xf86ReturnOptValBool(sna->Options, OPTION_ACCEL_DISABLE, FALSE) ||
 	    !sna_option_cast_to_bool(sna, OPTION_ACCEL_METHOD, TRUE)) {
@@ -484,7 +516,6 @@ static Bool sna_pre_init(ScrnInfoPtr scrn, int flags)
 	if (xf86ReturnOptValBool(sna->Options, OPTION_TILING_FB, FALSE))
 		sna->tiling &= ~SNA_TILING_FB;
 
-	sna->flags = 0;
 	if (!xf86ReturnOptValBool(sna->Options, OPTION_SWAPBUFFERS_WAIT, TRUE))
 		sna->flags |= SNA_NO_WAIT;
 	if (xf86ReturnOptValBool(sna->Options, OPTION_TRIPLE_BUFFER, TRUE))
@@ -551,6 +582,8 @@ sna_block_handler(BLOCKHANDLER_ARGS_DECL)
 
 	if (*tv == NULL || ((*tv)->tv_usec | (*tv)->tv_sec))
 		sna_accel_block_handler(sna, tv);
+
+	sna_xmir_post_damage(sna);
 }
 
 static void
@@ -629,6 +662,9 @@ sna_uevent_init(ScrnInfoPtr scrn)
 	Bool hotplug;
 	MessageType from = X_CONFIG;
 
+	if (sna->flags & SNA_IS_HOSTED)
+		return;
+
 	DBG(("%s\n", __FUNCTION__));
 
 	/* RandR will be disabled if Xinerama is active, and so generating
@@ -895,7 +931,10 @@ sna_screen_init(SCREEN_INIT_ARGS_DECL)
 	if (!miDCInitialize(screen, xf86GetPointerScreenFuncs()))
 		return FALSE;
 
-	if (xf86_cursors_init(screen, SNA_CURSOR_X, SNA_CURSOR_Y,
+	sna_xmir_init(sna, screen); /* after DirectCursor is initialised */
+
+	if ((sna->flags & SNA_IS_HOSTED) == 0 &&
+	    xf86_cursors_init(screen, SNA_CURSOR_X, SNA_CURSOR_Y,
 			       HARDWARE_CURSOR_TRUECOLOR_AT_8BPP |
 			       HARDWARE_CURSOR_BIT_ORDER_MSBFIRST |
 			       HARDWARE_CURSOR_INVERT_MASK |
@@ -988,7 +1027,10 @@ static Bool sna_enter_vt(VT_FUNC_ARGS_DECL)
 	struct sna *sna = to_sna(scrn);
 
 	DBG(("%s\n", __FUNCTION__));
-	if (!sna_become_master(sna))
+	if (intel_get_master(scrn))
+		return FALSE;
+
+	if (!sna_set_desired_mode(sna))
 		return FALSE;
 
 	if (sna->flags & SNA_REPROBE) {
@@ -1076,6 +1118,15 @@ static Bool sna_pm_event(SCRN_ARG_TYPE arg, pmEvent event, Bool undo)
 	return TRUE;
 }
 
+static Bool sna_enter_vt__hosted(VT_FUNC_ARGS_DECL)
+{
+	return TRUE;
+}
+
+static void sna_leave_vt__hosted(VT_FUNC_ARGS_DECL)
+{
+}
+
 Bool sna_init_scrn(ScrnInfoPtr scrn, int entity_num)
 {
 	DBG(("%s: entity_num=%d\n", __FUNCTION__, entity_num));
@@ -1106,13 +1157,18 @@ Bool sna_init_scrn(ScrnInfoPtr scrn, int entity_num)
 
 	scrn->PreInit = sna_pre_init;
 	scrn->ScreenInit = sna_screen_init;
-	scrn->SwitchMode = sna_switch_mode;
-	scrn->AdjustFrame = sna_adjust_frame;
-	scrn->EnterVT = sna_enter_vt;
-	scrn->LeaveVT = sna_leave_vt;
+	if (!hosted()) {
+		scrn->SwitchMode = sna_switch_mode;
+		scrn->AdjustFrame = sna_adjust_frame;
+		scrn->EnterVT = sna_enter_vt;
+		scrn->LeaveVT = sna_leave_vt;
+		scrn->ValidMode = sna_valid_mode;
+		scrn->PMEvent = sna_pm_event;
+	} else {
+		scrn->EnterVT = sna_enter_vt__hosted;
+		scrn->LeaveVT = sna_leave_vt__hosted;
+	}
 	scrn->FreeScreen = sna_free_screen;
-	scrn->ValidMode = sna_valid_mode;
-	scrn->PMEvent = sna_pm_event;
 
 	xf86SetEntitySharable(entity_num);
 	xf86SetEntityInstanceForScreen(scrn, entity_num,
